/// @file imgfilter.c
/// @brief Gaussian IMG filters.
///
#include "libtpcimgp.h"
/*****************************************************************************/

/*****************************************************************************/
/** Gaussian finite impulse response (FIR) isotropic or anisotropic image 
    filtering in 1-3 dimensions.

    Based on the code in:
    Getreuer P. A survey of Gaussian convolution algorithms.
    Image Processing On Line 2013;3:286–310. https://doi.org/10.5201/ipol.2013.87

    @sa imgGaussianAMFilter, imgSS
    @return If an error is encountered, function returns a non-zero value. Otherwise 0 is returned.
 */
int imgGaussianFIRFilter(
  /** Image data to be processed; data is overwritten with filtered image.
      All image frames are processed. */
  IMG *img,
  /** Gaussian S.D. in pixels in x dimension; SD=FWHM/2.355. Enter 0 to not filter in x dimension. */
  float xsd,
  /** Gaussian S.D. in pixels in y dimension; SD=FWHM/2.355. Enter 0 to not filter in y dimension. */
  float ysd,
  /** Gaussian S.D. in pixels in z dimension; SD=FWHM/2.355. Enter 0 to not filter in z dimension. */
  float zsd,
  /** Tolerance; enter 0.0 to use the default (1.0E-03). */
  double tolerance,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose
) {
  if(verbose>0) printf("%s(*img, %g, %g, %g, %g, %d)\n", __func__, xsd, ysd, zsd, tolerance, verbose); 

  if(img==NULL || img->dimz<1 || img->dimy<1 || img->dimx<1 || img->dimt<1) return 1;
  int dimx=img->dimx, dimy=img->dimy, dimz=img->dimz, dimt=img->dimt;
  if(dimx==1) xsd=0.0;
  if(dimy==1) ysd=0.0;
  if(dimz==1) zsd=0.0;
  if(xsd>0.0 && (dimx<4 || !(img->sizex>0.0))) return(2);
  if(ysd>0.0 && (dimy<4 || !(img->sizey>0.0))) return(3);
  if(zsd>0.0 && (dimz<4 || !(img->sizez>0.0))) return(4);
  if(!(xsd>0.0) && !(ysd>0.0) && !(zsd>0.0)) {
    if(verbose>1) printf("  no filtering done.\n");
    return(0);
  }
  if(!(tolerance>0.0)) tolerance=1.0E-03;
  if(tolerance>=1.0) return(5);

  double sqrt_pi2=sqrt(M_PI)/M_SQRT2; // sqrt(pi)/sqrt(2)

  /* 
   *  Filtering in x dimension
   */
  if(xsd>0.0) {
    if(verbose>1) {printf("  filtering in x dimension\n"); fflush(stdout);}
    int r=ceil(M_SQRT2*xsd*inverfc(0.5*tolerance)); // Radius of the filter
    double gtrunc[r+1]; // Truncated Gaussian filter for FIR convolution
    double sdsqrt2=M_SQRT2*xsd;
    gtrunc[0]=1.0;
//    for(int i=1; i<=r; i++) gtrunc[i]=exp(-0.5*((double)i/xsd)*((double)i/xsd));
    for(int i=1; i<=r; i++)
      gtrunc[i]=sqrt_pi2*xsd*(erf(((double)i+0.5)/sdsqrt2) - erf(((double)i-0.5)/sdsqrt2));
    double sum=gtrunc[0]; for(int i=1; i<=r; i++) sum+=2.0*gtrunc[i];
    for(int i=0; i<=r; i++) gtrunc[i]/=sum;
    if(verbose>2) printf("  r=%d \n", r);
    if(verbose>3) {
      printf("  gtrunc[]= %g", gtrunc[0]); for(int i=1; i<=r; i++) printf(",%g", gtrunc[i]);
      printf("\n");
    }

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      for(int z=0; z<dimz; z++) for(int y=0; y<dimy; y++) {
        double a[dimx];
        for(int x=0; x<dimx; x++) a[x]=img->m[z][y][x][t];
        for(int x=0; x<dimx; x++) {
          double v=gtrunc[0]*a[x];
          for(int m=1; m<=r; m++)
            v+=gtrunc[m]*(a[filterHSSBE(dimx, x-m)] + a[filterHSSBE(dimx, x+m)]);
          img->m[z][y][x][t]=v;
        }
      }
    }
  }


  /* 
   *  Filtering in y dimension
   */
  if(ysd>0.0) {
    if(verbose>1) {printf("  filtering in y dimension\n"); fflush(stdout);}
    int r=ceil(M_SQRT2*ysd*inverfc(0.5*tolerance)); // Radius of the filter
    double gtrunc[r+1]; // Truncated Gaussian filter for FIR convolution
    double sdsqrt2=M_SQRT2*ysd;
    gtrunc[0]=1.0;
//    for(int i=1; i<=r; i++) gtrunc[i]=exp(-0.5*((double)i/ysd)*((double)i/ysd));
    for(int i=1; i<=r; i++)
      gtrunc[i]=sqrt_pi2*ysd*(erf(((double)i+0.5)/sdsqrt2) - erf(((double)i-0.5)/sdsqrt2));
    double sum=gtrunc[0]; for(int i=1; i<=r; i++) sum+=2.0*gtrunc[i];
    for(int i=0; i<=r; i++) gtrunc[i]/=sum;
    if(verbose>2) printf("  r=%d \n", r);
    if(verbose>3) {
      printf("  gtrunc[]= %g", gtrunc[0]); for(int i=1; i<=r; i++) printf(",%g", gtrunc[i]);
      printf("\n");
    }

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      for(int z=0; z<dimz; z++) for(int x=0; x<dimx; x++) {
        double a[dimy];
        for(int y=0; y<dimy; y++) a[y]=img->m[z][y][x][t];
        for(int y=0; y<dimy; y++) {
          double v=gtrunc[0]*a[y];
          for(int m=1; m<=r; m++)
            v+=gtrunc[m]*(a[filterHSSBE(dimy, y-m)] + a[filterHSSBE(dimy, y+m)]);
          img->m[z][y][x][t]=v;
        }
      }
    }
  }


  /* 
   *  Filtering in z dimension
   */
  if(zsd>0.0) {
    if(verbose>1) {printf("  filtering in z dimension\n"); fflush(stdout);}
    int r=ceil(M_SQRT2*zsd*inverfc(0.5*tolerance)); // Radius of the filter
    double gtrunc[r+1]; // Truncated Gaussian filter for FIR convolution
    double sdsqrt2=M_SQRT2*zsd;
    gtrunc[0]=1.0;
//    for(int i=1; i<=r; i++) gtrunc[i]=exp(-0.5*((double)i/zsd)*((double)i/zsd));
    for(int i=1; i<=r; i++)
      gtrunc[i]=sqrt_pi2*zsd*(erf(((double)i+0.5)/sdsqrt2) - erf(((double)i-0.5)/sdsqrt2));
    double sum=gtrunc[0]; for(int i=1; i<=r; i++) sum+=2.0*gtrunc[i];
    for(int i=0; i<=r; i++) gtrunc[i]/=sum;
    if(verbose>2) printf("  r=%d \n", r);
    if(verbose>3) {
      printf("  gtrunc[]= %g", gtrunc[0]); for(int i=1; i<=r; i++) printf(",%g", gtrunc[i]);
      printf("\n");
    }

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      for(int y=0; y<dimy; y++) for(int x=0; x<dimx; x++) {
        double a[dimz];
        for(int z=0; z<dimz; z++) a[z]=img->m[z][y][x][t];
        for(int z=0; z<dimz; z++) {
          double v=gtrunc[0]*a[z];
          for(int m=1; m<=r; m++)
            v+=gtrunc[m]*(a[filterHSSBE(dimz, z-m)] + a[filterHSSBE(dimz, z+m)]);
          img->m[z][y][x][t]=v;
        }
      }
    }
  }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Alvarez-Mazorra approximate Gaussian image filtering in 1-3 dimensions.

    References:
    - Getreuer P. A survey of Gaussian convolution algorithms. https://doi.org/10.5201/ipol.2013.87
    - Alvarez L, Mazorra L, Signal and image restoration using shock Filters and anisotropic
      Diffusion. https://www.jstor.org/stable/2158018

    @sa imgGaussianFIRFilter, imgSS
    @return If an error is encountered, function returns a non-zero value. Otherwise 0 is returned.
 */
int imgGaussianAMFilter(
  /** Image data to be processed; data is overwritten with filtered image.
      All image frames are processed. */
  IMG *img,
  /** Gaussian S.D. in pixels in x dimension; SD=FWHM/2.355. Enter 0 to not filter in x dimension. */
  float xsd,
  /** Gaussian S.D. in pixels in y dimension; SD=FWHM/2.355. Enter 0 to not filter in y dimension. */
  float ysd,
  /** Gaussian S.D. in pixels in z dimension; SD=FWHM/2.355. Enter 0 to not filter in z dimension. */
  float zsd,
  /** Number of filtering passes, usually 4 or more; enter 0 to use the default steps (4). */
  int passNr,
  /** Tolerance; enter 0.0 to use the default (1.0E-06). */
  double tolerance,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose
) {
  if(verbose>0) printf("%s(*img, %g, %g, %g, %d, %d)\n", __func__, xsd, ysd, zsd, passNr, verbose); 

  if(img==NULL || img->dimz<1 || img->dimy<1 || img->dimx<1 || img->dimt<1) return 1;
  int dimx=img->dimx, dimy=img->dimy, dimz=img->dimz, dimt=img->dimt;
  if(dimx==1) xsd=0.0;
  if(dimy==1) ysd=0.0;
  if(dimz==1) zsd=0.0;
  if(xsd>0.0 && (dimx<4 || !(img->sizex>0.0))) return(2);
  if(ysd>0.0 && (dimy<4 || !(img->sizey>0.0))) return(3);
  if(zsd>0.0 && (dimz<4 || !(img->sizez>0.0))) return(4);
  if(!(xsd>0.0) && !(ysd>0.0) && !(zsd>0.0)) {
    if(verbose>1) printf("  no filtering done.\n");
    fflush(stdout); return(0);
  }
  if(passNr<=0) passNr=4;
  if(!(tolerance>0.0)) tolerance=1.0E-06;


  /* 
   *  Filtering in x dimension
   */
  if(xsd>0.0) {
    if(verbose>1) {printf("  filtering in x dimension\n"); fflush(stdout);}
    double q=xsd;
    if(0) { // Adjust SD for improved accuracy at low step numbers
      double f=0.7818+(double)passNr;
      q*= 1.0 + (0.3165*(double)passNr + 0.5695)/(f*f);
    }
    double lambda=(q*q)/(2.0*(double)passNr);
    double nu=(1.0 + 2.0*lambda - sqrt(1.0 + 4.0*lambda))/(2.0*lambda);
    double scale=pow(nu/lambda, passNr);
    int termNr=ceil(log((1.0-nu)*tolerance)/log(nu));
    if(verbose>2) printf("  nu=%g scale=%g termNr=%d\n", nu, scale, termNr);

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      double sum1=0.0, sum2=0.0;
      for(int z=0; z<dimz; z++) for(int y=0; y<dimy; y++) {
        double d[dimx];
        for(int x=0; x<dimx; x++) sum1+=d[x]=img->m[z][y][x][t];
        filterAMGaussian(d, dimx, passNr, nu, scale, termNr);
        for(int x=0; x<dimx; x++) sum2+=img->m[z][y][x][t]=d[x];
      }
      double f=sum1/sum2;
      if(verbose>1 && fabs(1.0-f)>1.0E-10) {
        printf("  frame %d: sum of voxels differs pre/post filtering %g/%g=%g\n", 1+t, sum1, sum2, f);
      }
    }
  }

  /* 
   *  Filtering in y dimension
   */
  if(ysd>0.0) {
    if(verbose>1) {printf("  filtering in y dimension\n"); fflush(stdout);}
    double q=ysd;
    if(0) { // Adjust SD for improved accuracy at low step numbers
      double f=0.7818+(double)passNr;
      q*= 1.0 + (0.3165*(double)passNr + 0.5695)/(f*f);
    }
    double lambda=(q*q)/(2.0*(double)passNr);
    double nu=(1.0 + 2.0*lambda - sqrt(1.0 + 4.0*lambda))/(2.0*lambda);
    double scale=pow(nu/lambda, passNr);
    int termNr=ceil(log((1.0-nu)*tolerance)/log(nu));
    if(verbose>2) printf("  nu=%g scale=%g termNr=%d\n", nu, scale, termNr);

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      double sum1=0.0, sum2=0.0;
      for(int z=0; z<dimz; z++) for(int x=0; x<dimx; x++) {
        double d[dimy];
        for(int y=0; y<dimy; y++) sum1+=d[y]=img->m[z][y][x][t];
        filterAMGaussian(d, dimy, passNr, nu, scale, termNr);
        for(int y=0; y<dimy; y++) sum2+=img->m[z][y][x][t]=d[y];
      }
      double f=sum1/sum2;
      if(verbose>1 && fabs(1.0-f)>1.0E-10) {
        printf("  frame %d: sum of voxels differs pre/post filtering %g/%g=%g\n", 1+t, sum1, sum2, f);
      }
    }
  }

  /* 
   *  Filtering in z dimension
   */
  if(zsd>0.0) {
    if(verbose>1) {printf("  filtering in z dimension\n"); fflush(stdout);}
    double q=xsd;
    if(0) { // Adjust SD for improved accuracy at low step numbers
      double f=0.7818+(double)passNr;
      q*= 1.0 + (0.3165*(double)passNr + 0.5695)/(f*f);
    }
    double lambda=(q*q)/(2.0*(double)passNr);
    double nu=(1.0 + 2.0*lambda - sqrt(1.0 + 4.0*lambda))/(2.0*lambda);
    double scale=pow(nu/lambda, passNr);
    int termNr=ceil(log((1.0-nu)*tolerance)/log(nu));
    if(verbose>2) printf("  nu=%g scale=%g termNr=%d\n", nu, scale, termNr);

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      double sum1=0.0, sum2=0.0;
      for(int y=0; y<dimy; y++) for(int x=0; x<dimy; x++) {
        double d[dimz];
        for(int z=0; z<dimz; z++) sum1+=d[z]=img->m[z][y][x][t];
        filterAMGaussian(d, dimz, passNr, nu, scale, termNr);
        for(int z=0; z<dimz; z++) sum2+=img->m[z][y][x][t]=d[z];
      }
      double f=sum1/sum2;
      if(verbose>1 && fabs(1.0-f)>1.0E-10) {
        printf("  frame %d: sum of voxels differs pre/post filtering %g/%g=%g\n", 1+t, sum1, sum2, f);
      }
    }
  }


  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** 1D Alvarez-Mazorra Gaussian filtering.
    @sa imgGaussianAMFilter
    @return If an error is encountered, function returns a non-zero value. Otherwise 0 is returned.
 */
int filterAMGaussian(
  /** Pointer to source data array, overwritten by the results. */
  double *a,
  /** Sample number. */
  int n,
  /** Number of filtering passes; usually 4. */
  int passNr,
  /** Filter coefficient. */
  double nu,
  /** Filter scale. */
  double scale,
  /** Number of terms needed to approximate the sum with required accuracy for the left boundary. */
  int termNr
) {
  if(a==NULL || n<3) return(1);
  if(!(nu>0.0)) return(2);
  if(passNr<1) passNr=4;
  if(termNr<6) termNr=6;

  /* Scale the data. */
  for(int i=0; i<n; i++) a[i]*=scale;

  /* Loop through the passes of filtering */
  for(int pass=0; pass<passNr; pass++) {
    a[0]=filterAMLB(a, n, nu, termNr);
    for(int i=1; i<n; i++) a[i]+=nu*a[i-1]; // causal filter
    a[n-1]/=(1.0-nu);
    for(int i=n-1; i>0; i--) a[i-1]+=nu*a[i]; // anti-causal filter
  }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** Half-sample symmetric boundary extension.
    @sa filterEbox, filterAMLB
    @return Reflected sample in array[0..n-1].
 */
inline int filterHSSBE(
  /** Sample number. */
  int n,
  /** Requested sample, possibly outside array[0..N-1]. */
  int r
) {
  while(1) {
    if(r<0) r=-1-r;
    else if(r>=n) r=2*n-1-r;
    else break;
  }
  return(r);
}
/*****************************************************************************/

/*****************************************************************************/
/** Left boundary for 1D Alvarez-Mazorra Gaussian filtering.
    @sa filterAMGaussian, filterHSSBE
    @return The sum approximating the first filtered sample value, or NaN in case of an error.
 */
double filterAMLB(
  /** Pointer to source data array; not modified. */
  double *a,
  /** Sample number. */
  int n,
  /** Filtering parameter. */
  double nu,
  /** Number of terms. */
  int termNr
) {
  if(a==NULL || n<1) return(nan(""));
  double h=1.0, sum=a[0];
  for(int m=1; m<termNr; m++) {
    h*=nu;
    sum+=h*a[filterHSSBE(n, -m)];
  }
  return(sum);
}
/*****************************************************************************/

/*****************************************************************************/
/** Gaussian extended box image filtering in 1-3 dimensions.

    References:
    - Getreuer P. A survey of Gaussian convolution algorithms.
      https://doi.org/10.5201/ipol.2013.87
    - Gwosdek et al. Theoretical foundations of Gaussian convolution by extended box filtering.
      https://doi.org/10.1007/978-3-642-24785-9_38
    - Elboher E, Werman M. Efficient and accurate Gaussian image filtering using running sums.
      https://arxiv.org/pdf/1107.4958.pdf
    - Wells WM. Efficient synthesis of Gaussian filters by cascaded uniform filters.
      https://doi.org/10.1109/TPAMI.1986.4767776

    @sa imgFast2DGaussianFilter, imgGaussianFilter, imgSS
    @return If an error is encountered, function returns a non-zero value. Otherwise 0 is returned.
 */
int imgGaussianEBoxFilter(
  /** Image data to be processed; data is overwritten with filtered image.
      All image frames are processed. */
  IMG *img,
  /** Gaussian S.D. in pixels in x dimension; SD=FWHM/2.355. Enter 0 to not filter in x dimension. */
  float xsd,
  /** Gaussian S.D. in pixels in y dimension; SD=FWHM/2.355. Enter 0 to not filter in y dimension. */
  float ysd,
  /** Gaussian S.D. in pixels in z dimension; SD=FWHM/2.355. Enter 0 to not filter in z dimension. */
  float zsd,
  /** Number of filtering passes, usually 3-5; enter 0 to use the default steps (4). */
  int passNr,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose
) {
  if(verbose>0) printf("%s(*img, %g, %g, %g, %d, %d)\n", __func__, xsd, ysd, zsd, passNr, verbose); 

  if(img==NULL || img->dimz<1 || img->dimy<1 || img->dimx<1 || img->dimt<1) return 1;
  int dimx=img->dimx, dimy=img->dimy, dimz=img->dimz, dimt=img->dimt;
  if(dimx==1) xsd=0.0;
  if(dimy==1) ysd=0.0;
  if(dimz==1) zsd=0.0;
  if(xsd>0.0 && (dimx<4 || !(img->sizex>0.0))) return(2);
  if(ysd>0.0 && (dimy<4 || !(img->sizey>0.0))) return(3);
  if(zsd>0.0 && (dimz<4 || !(img->sizez>0.0))) return(4);
  if(!(xsd>0.0) && !(ysd>0.0) && !(zsd>0.0)) {
    if(verbose>1) printf("  no filtering done.\n");
    return(0);
  }
  if(passNr<=0) passNr=4;


  /* 
   *  Filtering in x dimension
   */
  if(xsd>0.0) {
    if(verbose>1) {printf("  filtering in x dimension\n"); fflush(stdout);}
    int rib;
    double wob, wib, s2p=xsd*xsd/(double)passNr;
    /* Calculate inner box radius (by Wells) */
    rib=0.5*sqrt(1.0 + 12.0*s2p) - 0.5;
    if(verbose>3) printf("  precise_r := %f\n", 0.5*sqrt(1.0+12.0*s2p));
    /* Alpha */
    double a=(2*rib+1)*(rib*(rib+1) - 3.0*s2p) / (6.0*(s2p - (rib+1)*(rib+1)));
    /* Outer box weight */
    wob = a/(2.0*(a+rib)+1.0);
    /* Inner box weight */
    wib = 1.0-wob;
    if(verbose>2) {printf("  rib=%d wob=%g wib=%g a=%g\n", rib, wob, wib, a); fflush(stdout);}

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      double sum1=0.0, sum2=0.0;
      for(int z=0; z<dimz; z++) for(int y=0; y<dimy; y++) {
        double d[dimx];
        for(int x=0; x<dimx; x++) sum1+=d[x]=img->m[z][y][x][t];
        filterEbox(d, dimx, wob, wib, rib, passNr);
        for(int x=0; x<dimx; x++) sum2+=img->m[z][y][x][t]=d[x];
      }
      double f=sum1/sum2;
      if(fabs(1.0-f)>1.0E-10) {
        if(verbose>1)
          printf("  frame %d: sum of voxels differs pre/post filtering %g/%g=%g\n", 1+t, sum1, sum2, f);
        for(int z=0; z<dimz; z++) for(int y=0; y<dimy; y++) for(int x=0; x<dimx; x++)
          img->m[z][y][x][t]*=f;
      }
    }
  }

  /* 
   *  Filtering in y dimension
   */
  if(ysd>0.0) {
    if(verbose>1) {printf("  filtering in y dimension\n"); fflush(stdout);}
    int rib;
    double wob, wib, s2p=ysd*ysd/(double)passNr;
    /* Calculate inner box radius */
    rib=0.5*sqrt(1.0 + 12.0*s2p) - 0.5;
    /* Alpha */
    double a=(2*rib+1)*(rib*(rib+1) - 3.0*s2p) / (6.0*(s2p - (rib+1)*(rib+1)));
    /* Outer box weight */
    wob = a/(2.0*(a+rib)+1);
    /* Inner box weight */
    wib = 1.0-wob;
    if(verbose>2) {printf("  rib=%d wob=%g wib=%g a=%g\n", rib, wob, wib, a); fflush(stdout);}

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      double sum1=0.0, sum2=0.0;
      for(int z=0; z<dimz; z++) for(int x=0; x<dimx; x++) {
        double d[dimy];
        for(int y=0; y<dimy; y++) sum1+=d[y]=img->m[z][y][x][t];
        filterEbox(d, dimy, wob, wib, rib, passNr);
        for(int y=0; y<dimy; y++) sum2+=img->m[z][y][x][t]=d[y];
      }
      double f=sum1/sum2;
      if(fabs(1.0-f)>1.0E-10) {
        if(verbose>1)
          printf("  frame %d: sum of voxels differs pre/post filtering %g/%g=%g\n", 1+t, sum1, sum2, f);
        for(int z=0; z<dimz; z++) for(int y=0; y<dimy; y++) for(int x=0; x<dimx; x++)
          img->m[z][y][x][t]*=f;
      }
    }
  }

  /* 
   *  Filtering in z dimension
   */
  if(zsd>0.0) {
    if(verbose>1) {printf("  filtering in z dimension\n"); fflush(stdout);}
    int rib;
    double wob, wib, s2p=zsd*zsd/(double)passNr;
    /* Calculate inner box radius */
    rib=0.5*sqrt(1.0 + 12.0*s2p) - 0.5;
    /* Alpha */
    double a=(2*rib+1)*(rib*(rib+1) - 3.0*s2p) / (6.0*(s2p - (rib+1)*(rib+1)));
    /* Outer box weight */
    wob = a/(2.0*(a+rib)+1);
    /* Inner box weight */
    wib = 1.0-wob;
    if(verbose>2) {printf("  rib=%d wob=%g wib=%g a=%g\n", rib, wob, wib, a); fflush(stdout);}

    for(int t=0; t<dimt; t++) { // Image time frames are processed as separate images
      if(verbose>5) {printf("    frame %d\n", 1+t); fflush(stdout);}
      double sum1=0.0, sum2=0.0;
      for(int y=0; y<dimy; y++) for(int x=0; x<dimx; x++) {
        double d[dimz];
        for(int z=0; z<dimz; z++) sum1+=d[z]=img->m[z][y][x][t];
        filterEbox(d, dimz, wob, wib, rib, passNr);
        for(int z=0; z<dimz; z++) sum2+=img->m[z][y][x][t]=d[z];
      }
      double f=sum1/sum2;
      if(fabs(1.0-f)>1.0E-10) {
        if(verbose>1)
          printf("  frame %d: sum of voxels differs pre/post filtering %g/%g=%g\n", 1+t, sum1, sum2, f);
        for(int z=0; z<dimz; z++) for(int y=0; y<dimy; y++) for(int x=0; x<dimx; x++)
          img->m[z][y][x][t]*=f;
      }
    }
  }

  return(0);
}
/*****************************************************************************/

/*****************************************************************************/
/** 1D extended box filtering .
    @sa imgFast1DGaussianFilter
    @return If an error is encountered, function returns a non-zero value. Otherwise 0 is returned.
 */
int filterEbox(
  /** Pointer to source data array. */
  double *a,
  /** Sample number. */
  int n,
  /** Weight of the outer box. */
  double wob,
  /** Weight of the inner box. */
  double wib,
  /** Radius of the inner box. */
  int rib,
  /** Number of filtering passes. */
  int passNr
) {
  if(a==NULL || n<3 || wob<=0.0 || wib<=0.0 || rib<0) return(1);
  double buf[n];
  for(int pass=0; pass<passNr; pass++) {
    double accum=0.0;
    for(int i=-rib; i<=rib; i++) accum+=a[filterHSSBE(n, i)];
    buf[0]=wob*(a[filterHSSBE(n, rib+1)] + a[filterHSSBE(n, -rib-1)]) + (wob+wib)*accum;
    accum=buf[0];
    for(int i=1; i<n; i++) {
      accum += wob*(a[filterHSSBE(n, i+rib+1)] - a[filterHSSBE(n, i-rib-2)])
             + wib*(a[filterHSSBE(n, i+rib)] - a[filterHSSBE(n, i-rib-1)]);
      buf[i]=accum;
    }
    for(int i=0; i<n; i++) a[i]=buf[i];
  }
  return(0);
}
/*****************************************************************************/

/*****************************************************************************/





#if(0)










/*
*/







/*****************************************************************************/

/*****************************************************************************/
/** Make a 2D Gaussian convolution kernel for isotropic filtering.

    @sa imgConvolute2D, imgFast3DGaussianFilter, imgFast2DGaussianFilter, img2cube
    @return Returns 0 when successful, otherwise <>0.
*/
int imgFillGaussKernel(
  /** Gaussian convolution kernel matrix[size][size], filled here. */
  float **kernel,
  /** Gaussian S.D. in pixels (decimals are ok). */
  float stdev,
  /** Kernel dimension. */
  int size
) {
  int x, y;
  float mx, my, v, ksum=0.0;

  if(kernel==NULL || size<3) return 1;
  if(stdev<0.0) return 2;
  if((size%2)==0) return 3; // size must be odd number

  v=stdev*stdev;
  for(x=0; x<size; x++) {
    mx = x - size/2; // note: ints on purpose
    for(y=0; y<size; y++) {
      my = y - size/2; // note: ints on purpose
      if(stdev>0) {
#if(0)  // Gaussian in the middle of pixel is calculated.
        // Rousset and Zaidi (2006) https://doi.org/10.1007/0-387-25444-7_8
        // Brix et al (2002) Nuklearmedizin 2002;41:184-190.
        kernel[y][x] =
          (1.0/(2.0*M_PI*v)) * powf(M_E,-((mx*mx)+(my*my))/(2.0*v));
#else  // Pixel area is considered
        kernel[y][x] = 0.25
              *(erff((mx+0.5)/(stdev*M_SQRT2))-erff((mx-0.5)/(stdev*M_SQRT2)))
              *(erff((my+0.5)/(stdev*M_SQRT2))-erff((my-0.5)/(stdev*M_SQRT2)));
        // For 1D this is NOT
        // kernel[x]=(erff((mx+0.5)/(stdev*M_SQRT2))-erff((mx-0.5)/(stdev*M_SQRT2)))*(stdev/M_SQRT2);
#endif
      } else {
        if(x==size/2 && y==size/2) kernel[y][x]=1.0; else kernel[y][x]=0.0;
      }
      ksum+=kernel[y][x];
    }
  }

  /* Ensure quantitativity with normalization;
     divide each kernel value by their sum */
  v=1.0/ksum;
  for(x=0; x<size; x++) for(y=0; y<size; y++) kernel[y][x]*=v;

  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Perform a convolution operation on one image matrix.
    @remark Note that this may be a very slow process if image and kernel size is
    large: width*height*size*size multiplications will be done.

    @sa imgFast3DGaussianFilter, imgFast2DGaussianFilter
    @return Returns 0 when successful, otherwise <>0.
 */
int imgConvolute2D(
  /** Image matrix data[height][width][frame]. */
  float ***data,
  /** Temporary preallocated memory buffer[size][width+size-1]. */
  float **buffer,
  /** frame [0..dimt] of data matrix which is processed. */
  int frame,
  /** Width of image matrix (dimx). */
  int width,
  /** Height of image matrix (dimy). */
  int height,
  /** Convolution kernel[size][size] matrix. */
  float **kernel,
  /** Convolution kernel size; must be an odd number, and >=3, and
      smaller than 1 + 2*image_dimension. */
  int size,
  /** Fill borders with zero (0) or with closest image pixel (<>0). */
  int border,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose,
  /** Pointer to error message, at least 128 characters; NULL, if not needed. */
  char *errmsg 
) {
  const int size2=size/2, bufw=width+2*size2;
  int x, y, i;
  float *tmp;
  float filtered;
  int kx, ky;

  if(verbose>0) 
    printf("%s(***data, **buf, %d, %d, %d, **ker, %d, %d, %d)\n", 
           __func__, frame, width, height, size, border, verbose);
  /* Check input */
  if(errmsg!=NULL) strcpy(errmsg, "invalid function input");
  if(data==NULL || buffer==NULL || frame<0 || width<3 || height<3 || kernel==NULL)
    return 1;
  if((size%2)==0) {
    if(errmsg!=NULL) strcpy(errmsg, "kernel size is even number");
    return 1;
  }
  if(size<3) {
    if(errmsg!=NULL) strcpy(errmsg, "kernel size must be >=3");
    return 1;
  } else {
    i=width<height ? width : height;
    if(size>=2*i+1) {
      if(errmsg!=NULL) strcpy(errmsg, "kernel size is too big for image");
      return 1;
    }
  }
  if(verbose>1) {
    printf("  size2 := %d\n", size2);
    printf("  bufw := %d\n", bufw);
  }

  /* Fill the initial buffer */
  if(verbose>2) printf("filling initial data buffer\n"); 
  if(border==0) {
    for(y=0; y<size2; y++) {
      for(x=0; x<bufw; x++) buffer[y][x]=0;
    }
    for(y=size2; y<size; y++) {
      for(x=0; x<size2; x++) buffer[y][x]=0;
      for(x=size2; x<bufw-size2; x++) buffer[y][x]=data[y-size2][x-size2][frame];
      for(x=bufw-size2; x<bufw; x++) buffer[y][x]=0;
    }
  } else {
    for(y=0; y<size2; y++) {
      for(x=0; x<size2; x++) buffer[y][x]=data[0][0][frame];
      for(x=size2; x<bufw-size2; x++) buffer[y][x]=data[0][x-size2][frame];
      for(x=bufw-size2; x<bufw; x++) buffer[y][x]=data[0][width-1][frame];
    }
    for(y=size2; y<size; y++) {
      for(x=0; x<size2; x++) buffer[y][x]=data[0][y-size2][frame];
      for(x=size2; x<bufw-size2; x++) buffer[y][x]=data[y-size2][x-size2][frame];
      for(x=bufw-size2; x<bufw; x++) buffer[y][x]=data[y-size2][width-1][frame];
    }
  }
  if(verbose>8) {
    int xi, yi;
    printf("Initial buffer for row-index 0:\n");
    for(yi=0; yi<size; yi++) {
      for(xi=0; xi<bufw; xi++) printf(" %4.0f", buffer[yi][xi]);
      printf("\n");
    }
  }

  /* Filter */
  for(y=0; y<height; y++) {
    if(verbose>2) printf("filtering image row %d\n", y+1); 

    /* Filter a row */
    for(x=0; x<width; x++) {
      filtered = 0.0;
      for(ky=0; ky<size; ky++) {
        for(kx=0; kx<size; kx++) {
          filtered += buffer[ky][x+kx] * kernel[ky][kx];
        }
      }
      data[y][x][frame] = filtered;
    }

    if(y==height-1) break; // no need for further buffer filling

    if(verbose>2) printf("filling data buffer for next image row\n"); 
    /* Shift rows in data buffer */
    tmp = buffer[0];
    for(i=1; i<size; i++) buffer[i-1]=buffer[i];
    buffer[size-1]=tmp;

    /* Get next row from image or 'guess' it */
    if(y+size2+1<height) { // still inside original image 
      if(border==0) {
        for(x=0; x<size2; x++)
          buffer[size-1][x]=0;
        for(x=size2; x<bufw-size2; x++)
          buffer[size-1][x]=data[y+size2+1][x-size2][frame];
        for(x=bufw-size2; x<bufw; x++)
          buffer[size-1][x]=0;
      } else {
        for(x=0; x<size2; x++)
          buffer[size-1][x]=data[y+size2+1][0][frame];
        for(x=size2; x<bufw-size2; x++)
          buffer[size-1][x]=data[y+size2+1][x-size2][frame];
        for(x=bufw-size2; x<bufw; x++)
          buffer[size-1][x]=data[y+size2+1][width-1][frame];
      }
    } else { // below original image
      if(border==0) {
        for(x=0; x<bufw; x++) buffer[size-1][x] = 0;
      } else {
        for(x=0; x<size2; x++)
          buffer[size-1][x]=data[height-1][0][frame];
        for(x=size2; x<bufw-size2; x++)
          buffer[size-1][x]=data[height-1][x-size2][frame];
        for(x=bufw-size2; x<bufw; x++)
          buffer[size-1][x]=data[height-1][width-1][frame];
      }
    }
    if(verbose>9) {
      int xi, yi;
      printf("Buffer for row-index %d:\n", y+1);
      for(yi=0; yi<size; yi++) {
        for(xi=0; xi<bufw; xi++) printf(" %4.0f", buffer[yi][xi]);
        printf("\n");
      }
    }
  }
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Apply 2D Gaussian filter to whole dynamic image in IMG structure.
    @sa imgFast3DGaussianFilter, imgFast2DGaussianFilter, imgsegmSimilar, imgSwell
    @return If an error is encountered, function returns a non-zero value. Otherwise 0 is returned.
 */
int imgGaussianFilter(
  /** Image data to be processed; data is overwritten with filtered image. */
  IMG *img,
  /** Plane index [0..dimz-1]; enter <0 to filter all image planes. */
  int plane,
  /** Frame index [0..dimt-1]; enter <0 to filter all image time frames. */
  int frame,
  /** Gaussian S.D. in pixels (decimals are ok). */
  float gauss_sd,
  /** Gaussian kernel size in pixels; it must be an odd number and at least 3
      and smaller than 2*imgdim+1; typically 6*gauss_sd is sufficient;
      enter zero to set it automatically to 6*gauss_sd (plus 1 if necessary). */
  int size,
  /** In filtering, fill borders with zero (0) or with closest image pixel (<>0). */
  int border,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose,
  /** Pointer to error message, at least 128 characters; NULL, if not needed. */
  char *errmsg 
) {
  int i, j, zi, fi, mindim, ret;
  float **buffer, *bdata;
  float **gauss, *gdata;

  if(verbose>0)
    printf("%s(*img, %d, %d, %g, %d, %d, %d, msg)\n",
           __func__, plane, frame, gauss_sd, size, border, verbose); 

  /* Check input */
  if(errmsg!=NULL) strcpy(errmsg, "invalid function input");
  /* x and y dim must be >=2 so that kernel size can be at least 3 */
  if(img==NULL || img->dimz<1) return 1;
  if(plane>=0 && plane>=img->dimz) return 2;
  if(frame>=0 && frame>=img->dimt) return 3;
  mindim=img->dimx<img->dimy ? img->dimx : img->dimy;
  if(mindim<3) return 1;
  if(gauss_sd<0.0) return 2; // SD=0 is accepted for basic testing
  if(size!=0 && (size<3 || (size%2)==0) ) return 4;
  if(size==0) {
    size=6*roundf(gauss_sd); if(size<3) size=3;
    if(size>=2*mindim+1) size=2*mindim-1; // limit kernel size
    if((size%2)==0) size++; // make sure that it is odd number
    if(verbose>0) printf("size := %d\n", size);
  }

  /* Allocate memory for the kernel */
  if(errmsg!=NULL) strcpy(errmsg, "cannot allocate memory for kernel");
  if(verbose>2) printf("allocating memory for the kernel\n");
  gdata=(float*)malloc( size*size*sizeof(float) );
  if(gdata==NULL) {return 7;}
  gauss=(float**)malloc(size*sizeof(float*));
  if(gauss==NULL) {free(gdata); return 7;}
  for(i=0; i<size; i++) gauss[i]=gdata+(i*size);

  /* Allocate memory for temporary buffer[size][width+size-1] */
  if(errmsg!=NULL) strcpy(errmsg, "cannot allocate memory for data buffer");
  if(verbose>2) printf("allocating memory for the raw buffer\n");
  bdata=(float*)malloc( (img->dimx+size-1)*size*sizeof(float) );
  if(bdata==NULL) {free(gauss); free(gdata); return 8;}
  if(verbose>2) printf("allocating memory for the buffer\n");
  buffer=(float**)malloc(size*sizeof(float*));
  if(buffer==NULL) {free(gauss); free(gdata); free(bdata); return 8;}
  if(verbose>2) printf("preparing buffer\n");
  for(i=0; i<size; i++) buffer[i]=bdata+(i*(img->dimx+size-1));

  /* Make a gaussian convolution kernel */
  if(verbose>1) printf("calculating Gaussian convolution kernel\n");
  if(imgFillGaussKernel(gauss, gauss_sd, size) != 0) {
    if(errmsg!=NULL) strcpy(errmsg, "cannot compute Gaussian kernel");
    free(gauss); free(gdata); free(buffer); free(bdata);
    return 9;
  }
  if(verbose>4) {
    printf("Gaussian convolution kernel:\n");
    for(i=0; i<size; i++) {
      printf(" ");
      for(j=0; j<size; j++) printf(" %8.6f", gauss[i][j]);
      printf("\n");
    }
  }

  /* Convolute each image matrix */
  if(verbose>0) printf("convolution of the image data\n");
  for(zi=0; zi<img->dimz; zi++) {
    if(plane>=0 && plane!=zi) continue;
    if(verbose>1) printf("  plane %d\n", zi+1);
    for(fi=0; fi<img->dimt; fi++) {
      if(frame>=0 && frame!=fi) continue;
      if(verbose>1) printf("    frame %d\n", fi+1);
      ret=imgConvolute2D(img->m[zi], buffer, fi, img->dimx, img->dimy,
                         gauss, size, border, verbose, errmsg);
      if(ret!=0) {
        free(gauss); free(gdata); free(buffer); free(bdata);
        return(10+ret);
      }
    }
  }

  /* Free memory */
  free(gauss); free(gdata); free(buffer); free(bdata);

  if(verbose>1) printf("  convolution done.\n");
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Apply fast approximate 2D Gaussian filter to whole dynamic image in IMG struct.

    This function implements the fast Gaussian convolution algorithm with IIR approximation 
    (Alvarez L and Mazorra L, SIAM Journal on Numerical Analysis, 1994;31(2):590-605), and 
    is based on C code written by Pascal Getreuer <http://www.getreuer.info/home/gaussianiir>.

    @sa imgFast3DGaussianFilter, imgsegmSimilar, imgSwell, imgScale, imgGaussianFilter
    @return If an error is encountered, function returns a non-zero value. 
        Otherwise 0 is returned.
 */
int imgFast2DGaussianFilter(
  /** Image data to be processed; data is overwritten with filtered image. */
  IMG *img,
  /** Plane index [0..dimz-1]; enter <0 to filter all image planes. */
  int plane,
  /** Frame index [0..dimt-1]; enter <0 to filter all image time frames. */
  int frame,
  /** Gaussian S.D. in pixels (decimals are ok); SD=FWHM/2.355. */
  float gauss_sd,
  /** Number of time steps. More steps implies better accuracy and longer execution time; 
      enter 0 to use the default steps (4); too many steps may lead to floating-point problems. */
  int step_nr,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose,
  /** Pointer to error message, at least 128 characters; NULL, if not needed. */
  char *errmsg 
) {
  int i, xi, yi, zi, fi, mindim, step;
  double lambda, nu, boundaryscale, postscale;
  double *ptr, *dimg;


  if(verbose>0)
    printf("%s(*img, %d, %d, %g, %d, %d, msg)\n", __func__, plane, frame, gauss_sd, step_nr, verbose); 

  /* Check input */
  if(errmsg!=NULL) strcpy(errmsg, "invalid function input");
  if(gauss_sd==0.0) return 0;
  if(img==NULL || img->dimz<1) return 1;
  if(plane>=0 && plane>=img->dimz) return 2;
  if(frame>=0 && frame>=img->dimt) return 3;
  mindim=img->dimx<img->dimy ? img->dimx : img->dimy;
  if(mindim<3) return 1;
  if(gauss_sd<0.0) return 4; // SD=0 is accepted for basic testing
  /* Check step_nr */
  if(step_nr<=0) step_nr=4;

  /* Prepare for filtering */
  lambda= (gauss_sd*gauss_sd)/(2.0*step_nr);
  nu= (1.0 + 2.0*lambda - sqrt(1.0 + 4.0*lambda))/(2.0*lambda);
  boundaryscale= 1.0/(1.0-nu);
  postscale = pow(nu/lambda, 2*step_nr);
  if(verbose>1) {
    printf("nu := %g\n", nu);
    printf("boundaryscale := %g\n", boundaryscale);
    printf("postscale := %g\n", postscale);
  }

  /* Setup memory for a image plane */
  if(verbose>1) printf("Allocating memory\n");
  dimg=(double*)malloc(img->dimx*img->dimy*sizeof(double));
  if(dimg==NULL) {if(errmsg!=NULL) strcpy(errmsg, "out of memory"); return 5;}

  /* Process the required image planes and frames */
  if(verbose>1) printf("Gaussian filtering...\n");
  for(zi=0; zi<img->dimz; zi++) {
    if(plane>=0 && plane!=zi) continue;
    if(verbose>2) printf("  plane %d\n", zi+1);
    for(fi=0; fi<img->dimt; fi++) {
      if(frame>=0 && frame!=fi) continue;
      if(verbose>2) printf("    frame %d\n", fi+1);
      /* Copy image plane data to double buffer */
      for(yi=i=0; yi<img->dimy; yi++)
        for(xi=0; xi<img->dimx; xi++)
          dimg[i++]=img->m[zi][yi][xi][fi];
      /* Filter horizontally along each image row */
      for(yi=0; yi<img->dimy; yi++) {
        for(step=0; step<step_nr; step++) {
          ptr=dimg+img->dimx*yi; ptr[0]*=boundaryscale;
          /* Filter rightwards */
          for(xi=1; xi<img->dimx; xi++) ptr[xi]+=nu*ptr[xi-1];
          ptr[xi=img->dimx-1]*=boundaryscale;
          /* Filter leftwards */
          for(; xi>0; xi--) ptr[xi-1]+=nu*ptr[xi];
        }
      }
      /* Filter vertically along each image column */
      for(xi=0; xi<img->dimx; xi++) {
        for(step=0; step<step_nr; step++) {
          ptr=dimg+xi; ptr[0]*=boundaryscale;
          /* Filter downwards */
          for(i=img->dimx; i<img->dimx*img->dimy; i+=img->dimx)
            ptr[i]+=nu*ptr[i-img->dimx];
          ptr[i=img->dimx*(img->dimy-1)]*=boundaryscale;
          /* Filter upwards */
          for(; i>0; i-=img->dimx) ptr[i-img->dimx]+=nu*ptr[i];
        }
      }
      /* Copy and scale filtered image plane data back from buffer */
      for(yi=i=0; yi<img->dimy; yi++)
        for(xi=0; xi<img->dimx; xi++)
          img->m[zi][yi][xi][fi]=postscale*dimg[i++];
    } // next frame
  } // next plane

  /* Free memory */
  free(dimg);

  if(errmsg!=NULL) strcpy(errmsg, "ok");
  if(verbose>1) printf("  Gaussian convolution done.\n");
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Apply fast approximate 3D Gaussian filter to whole dynamic image in IMG structure.

    This function implements the fast Gaussian convolution algorithm with IIR approximation 
    (Alvarez L and Mazorra L, SIAM Journal on Numerical Analysis, 1994;31(2):590-605), and 
    is based on C code written by Pascal Getreuer <http://www.getreuer.info/home/gaussianiir>.

    @sa imgFast2DGaussianFilter, imgThresholdingLowHigh, imgSmoothMax, imgGaussianFilter, imgSS
    @return If an error is encountered, function returns a non-zero value. 
        Otherwise 0 is returned.
 */
int imgFast3DGaussianFilter(
  /** Image data to be processed; data is overwritten with filtered image;
      image must contain at least 3 planes. */
  IMG *img,
  /** Frame index [0..dimt-1]; enter <0 to filter all image time frames. */
  int frame,
  /** Gaussian S.D. in pixels (decimals are ok); note that the same S.D. is assumed to apply to 
      all dimensions, including Z dimension. SD=FWHM/2.355. */
  float gauss_sd,
  /** Number of time steps. More steps implies better accuracy and longer execution time; 
      enter 0 to use the default steps (4); too many steps may lead to floating-point problems. */
  int step_nr,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose,
  /** Pointer to error message, at least 128 characters; NULL, if not needed. */
  char *errmsg 
) {
  int i, xi, yi, zi, fi, mindim, step;
  double lambda, nu, boundaryscale, postscale;
  double *ptr, *dimg;


  if(verbose>0) printf("%s(*img, %d, %g, %d, %d, msg)\n", __func__, frame, gauss_sd, step_nr, verbose); 

  /* Check input */
  if(errmsg!=NULL) strcpy(errmsg, "invalid function input");
  if(gauss_sd==0.0) return 0; // SD=0 is accepted for basic testing
  if(img==NULL || img->dimz<3) return 1;
  if(frame>=0 && frame>=img->dimt) return 3;
  mindim=img->dimx<img->dimy ? img->dimx : img->dimy;
  if(mindim<3) return 1;
  if(!(gauss_sd>0.0)) return 4;
  /* Check step_nr */
  if(step_nr<=0) step_nr=4;

  /* Prepare for filtering */
  lambda= (gauss_sd*gauss_sd)/(2.0*step_nr);
  nu= (1.0 + 2.0*lambda - sqrt(1.0 + 4.0*lambda))/(2.0*lambda);
  boundaryscale= 1.0/(1.0-nu);
  postscale = pow(nu/lambda, 3*step_nr);
  if(verbose>1) {
    printf("nu := %g\n", nu);
    printf("boundaryscale := %g\n", boundaryscale);
    printf("postscale := %g\n", postscale);
  }

  /* Setup memory for a image frame (3D matrix) */
  if(verbose>1) printf("Allocating memory\n");
  dimg=(double*)malloc(img->dimz*img->dimx*img->dimy*sizeof(double));
  if(dimg==NULL) {if(errmsg!=NULL) strcpy(errmsg, "out of memory"); return 5;}

  /* Process the required image frame(s) */
  if(verbose>1) printf("Gaussian filtering...\n");
  for(fi=0; fi<img->dimt; fi++) {
    if(frame>=0 && frame!=fi) continue;
    if(verbose>2) printf("  frame %d\n", fi+1);
    /* Copy image 3D matrix data to double buffer */
    for(zi=i=0; zi<img->dimz; zi++)
      for(yi=0; yi<img->dimy; yi++)
        for(xi=0; xi<img->dimx; xi++)
          dimg[i++]=img->m[zi][yi][xi][fi];
    /* Filter horizontally along each image row */
    for(zi=0; zi<img->dimz; zi++) {
      for(yi=0; yi<img->dimy; yi++) {
        for(step=0; step<step_nr; step++) {
          ptr = dimg+img->dimx*(yi+img->dimy*zi); ptr[0]*=boundaryscale;
          /* Filter rightwards */
          for(xi=1; xi<img->dimx; xi++) ptr[xi]+=nu*ptr[xi-1];
          ptr[xi=img->dimx-1]*=boundaryscale;
          /* Filter leftwards */
          for(; xi>0; xi--) ptr[xi-1]+=nu*ptr[xi];
        }
      }
    }
    /* Filter vertically along each image column */
    for(zi=0; zi<img->dimz; zi++) {
      for(xi=0; xi<img->dimx; xi++) {
        for(step=0; step<step_nr; step++) {
          ptr=dimg+xi+img->dimx*img->dimy*zi; ptr[0]*=boundaryscale;
          /* Filter downwards */
          for(i=img->dimx; i<img->dimx*img->dimy; i+=img->dimx)
            ptr[i]+=nu*ptr[i-img->dimx];
          ptr[i=img->dimx*(img->dimy-1)]*=boundaryscale;
          /* Filter upwards */
          for(; i>0; i-=img->dimx) ptr[i-img->dimx]+=nu*ptr[i];
        }
      }
    }
    /* Filter along image z-dimension */
    for(yi=0; yi<img->dimy; yi++) {
      for(xi=0; xi<img->dimx; xi++) {
        for(step=0; step<step_nr; step++) {
          ptr=dimg+xi+img->dimx*yi; ptr[0]*=boundaryscale;
          for(i=img->dimx*img->dimy; i<img->dimx*img->dimy*img->dimz; i+=img->dimx*img->dimy)
            ptr[i]+=nu*ptr[i-img->dimx*img->dimy];
          ptr[i=img->dimx*img->dimy*(img->dimz-1)]*=boundaryscale;
          for(; i>0; i-=img->dimx*img->dimy)
            ptr[i-img->dimx*img->dimy]+=nu*ptr[i];
        }
      }
    }
    /* Copy and scale image 3D matrix data back from the buffer */
    for(zi=i=0; zi<img->dimz; zi++)
      for(yi=0; yi<img->dimy; yi++)
        for(xi=0; xi<img->dimx; xi++)
          img->m[zi][yi][xi][fi]=postscale*dimg[i++];
  } // next frame

  /* Free memory */
  free(dimg);

  if(errmsg!=NULL) strcpy(errmsg, "ok");
  if(verbose>1) printf("  Gaussian convolution done.\n");
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Apply fast approximate 1D Gaussian filter over planes (z-axis) to whole dynamic image 
    in IMG structure.

    This function implements the fast Gaussian convolution algorithm with IIR approximation 
    (Alvarez L and Mazorra L, SIAM Journal on Numerical Analysis, 1994;31(2):590-605), and 
    is based on C code written by Pascal Getreuer <http://www.getreuer.info/home/gaussianiir>.

    @sa imgFast3DGaussianFilter, imgFast2DGaussianFilter, imgGaussianFilter
    @return If an error is encountered, function returns a non-zero value. 
        Otherwise 0 is returned.
 */
int imgFast1DGaussianFilter(
  /** Image data to be processed; data is overwritten with filtered image;
      Plane number (z-dimension) must be at least 3. */
  IMG *img,
  /** Gaussian S.D. in pixels (decimals are ok); SD=FWHM/2.355. */
  float gauss_sd,
  /** Number of time steps. More steps implies better accuracy and longer execution time; 
      enter 0 to use the default steps (4); too many steps may lead to floating-point problems. */
  int step_nr,
  /** Verbose level; if zero, then only warnings are printed into stderr. */
  int verbose,
  /** Pointer to error message, at least 128 characters; NULL, if not needed. */
  char *errmsg 
) {
  double lambda, nu, boundaryscale, postscale;


  if(verbose>0)
    printf("%s(*img, %g, %d, %d, msg)\n", __func__, gauss_sd, step_nr, verbose);

  /* Check input */
  if(errmsg!=NULL) strcpy(errmsg, "invalid function input");
  if(gauss_sd==0.0) return 0;
  if(img==NULL || img->dimz<3) return 1;
  if(gauss_sd<0.0) return 4; // SD=0 is accepted for basic testing
  /* Check step_nr */
  if(step_nr<=0) step_nr=4;

  /* Prepare for filtering */
  lambda= (gauss_sd*gauss_sd)/(2.0*step_nr);
  nu= (1.0 + 2.0*lambda - sqrt(1.0 + 4.0*lambda))/(2.0*lambda);
  boundaryscale= 1.0/(1.0-nu);
  postscale = pow(nu/lambda, step_nr);
  if(verbose>1) {
    printf("nu := %g\n", nu);
    printf("boundaryscale := %g\n", boundaryscale);
    printf("postscale := %g\n", postscale);
  }

  /* Set-up memory for an array of pixel values across planes */
  if(verbose>1) printf("Allocating memory\n");
  double *data=(double*)malloc(img->dimz*sizeof(double));
  if(data==NULL) {if(errmsg!=NULL) strcpy(errmsg, "out of memory"); return 5;}

  /* Process */
  if(verbose>1) printf("Gaussian filtering...\n");
  for(int yi=0; yi<img->dimy; yi++) {
    for(int xi=0; xi<img->dimx; xi++) {
      for(int fi=0; fi<img->dimt; fi++) {
        /* Copy image data to double buffer */
        for(int zi=0; zi<img->dimz; zi++) data[zi]=img->m[zi][yi][xi][fi];
        /* 1D Gaussian filtering */
        for(int si=0; si<step_nr; si++) {
          int i=0;
          data[i]*=boundaryscale;
          /* Filter rightwards (causal) */  
          for(i=1; i<img->dimz; i++) data[i]+=nu*data[i-1];
          i=img->dimz-1; data[i]*=boundaryscale;
          /* Filter leftwards (anti-causal) */  
          for(; i>0; i--) data[i-1]+=nu*data[i];  
        }
        for(int zi=0; zi<img->dimz; zi++) data[zi]*=postscale;
        /* Copy processed buffer data back to image */
        for(int zi=0; zi<img->dimz; zi++) img->m[zi][yi][xi][fi]=data[zi];
      } // next frame
    } // next column
  } // next row

  /* Free memory */
  free(data);

  if(errmsg!=NULL) strcpy(errmsg, "ok");
  if(verbose>1) printf("  Gaussian 1D filtering done.\n");
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
#endif
