/** @file simdispersion.c
 *  @brief Simulation of dispersion.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpccm.h"
/*****************************************************************************/

/*****************************************************************************/
/** Simulate the effect of dispersion on a time-activity curve.

    The units of rate constants must be related to the TAC time units;
    1/min and min, or 1/sec and sec.
   
    @return Function returns 0 when successful, else a value >= 1.
    @author Vesa Oikonen
    @sa simC1, corDispersion
 */
int simDispersion(
  /** Array of sample times; must be in increasing order. */
  double *x,
  /** Array of sample values, which will be replaced here by dispersion added values. */
  double *y,
  /** Nr of samples. */
  const int n,
  /** First dispersion time constant (zero if no dispersion); in same time unit as sample times. */ 
  const double tau1,
  /** 2nd dispersion time constant (zero if no dispersion); in same time unit as sample times. */ 
  const double tau2,
  /** Array for temporary data, for at least n samples; enter NULL to let function to 
      allocate and free the temporary space. */
  double *tmp
) {
  /* Check input */
  if(x==NULL || y==NULL || n<2) return 1;
  if(tau1<0.0 || tau2<0.0) return 2;

  /* Allocate memory if not allocated by user */
  double *buf;
  if(tmp!=NULL) buf=tmp; else buf=(double*)malloc(n*sizeof(double));
  if(buf==NULL) return 3;

  /* First dispersion */
  if(tau1>0.0) {
    double k=1.0/tau1;
    int ret=simC1(x, y, n, k, k, buf); 
    if(ret!=0) {
      if(tmp==NULL) free(buf);
      return 100+ret;
    }
    for(int i=0; i<n; i++) y[i]=buf[i];
  }

  /* Second dispersion */
  if(tau2>0.0) {
    double k=1.0/tau2;
    int ret=simC1(x, y, n, k, k, buf);
    if(ret!=0) {
      if(tmp==NULL) free(buf);
      return 200+ret;
    }
    for(int i=0; i<n; i++) y[i]=buf[i];
  }

  if(tmp==NULL) free(buf);
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
/** Correct time-activity curve for the effect of dispersion.

    Data must be noise-free and have very short sampling intervals.
    The units of rate constants must be related to the TAC time units;
    1/min and min, or 1/sec and sec.
   
    @return Function returns 0 when successful, else a value >= 1.
    @author Vesa Oikonen
    @sa simDispersion
 */
int corDispersion(
  /** Array of sample times; must be in increasing order and >=0. */
  double *x,
  /** Array of sample values, which will be replaced here by dispersion corrected values. */
  double *y,
  /** Nr of samples. */
  const int n,
  /** Dispersion time constant (zero if no dispersion); in same time unit as sample times. */ 
  const double tau,
  /** Array for temporary data, for at least n samples; enter NULL to let function to 
      allocate and free the temporary space. */
  double *tmp
) {
  /* Check input */
  if(x==NULL || y==NULL || n<2) return 1;
  if(tau<0.0) return 2;
  if(x[0]<0.0) return 3;

  /* Allocate memory if not allocated by user */
  double *buf;
  if(tmp!=NULL) buf=tmp; else buf=(double*)malloc(n*sizeof(double));
  if(buf==NULL) return 3;

  /* Integrate measured data */
  buf[0]=0.5*y[0]*x[0];
  for(int i=1; i<n; i++) buf[i]=buf[i-1]+0.5*(y[i]+y[i-1])*(x[i]-x[i-1]);

  /* Calculate the integral of dispersion corrected data */
  for(int i=0; i<n; i++) buf[i]+=tau*y[i];

  /* Calculate dispersion corrected curve from its integral */
  {
    int i=0;
    double dt=x[i];
    if(dt>0.0) y[i]=2.0*buf[i]/dt; else y[i]=0.0;
    for(i=1; i<n-1; i++) {
      dt=x[i+1]-x[i-1]; if(dt>0.0) y[i]=(buf[i+1]-buf[i-1])/dt; else y[i]=y[i-1];
    }
    dt=x[i]-x[i-1]; if(dt>0.0) y[i]=(buf[i]-buf[i-1])/dt; else y[i]=y[i-1];
  }

  if(tmp==NULL) free(buf);
  return 0;
}
/*****************************************************************************/

/*****************************************************************************/
