/// @file tpcextensions.h
/// @brief Header file for library libtpcextensions.
/// @author Vesa Oikonen
/// @copyright (c) Turku PET Centre
///
#ifndef _TPCEXTENSIONS_H_
#define _TPCEXTENSIONS_H_
/*****************************************************************************/

/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <errno.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <math.h>
#include <float.h>
#include <time.h>
/*****************************************************************************/

/*****************************************************************************/
/// Define max units string length
#ifndef MAX_UNITS_LEN
#define MAX_UNITS_LEN 32 ///<  Max string length for PET units
#endif
/*****************************************************************************/

/*****************************************************************************/
/// Max length of TAC ID name (not including trailing zero)
#ifndef MAX_TACNAME_LEN
#define MAX_TACNAME_LEN 31 ///<  Max string length for TAC name
#endif
/** Max length of parameter names and units */
#ifndef MAX_PARNAME_LEN
#define MAX_PARNAME_LEN MAX_TACNAME_LEN ///<  Max string length for PAR name
#endif
/*****************************************************************************/

/*****************************************************************************/
/// Define max study nember length
#ifndef MAX_STUDYNR_LEN
#define MAX_STUDYNR_LEN 255 ///<  Max string length for PET study number
#endif
/*****************************************************************************/

/*****************************************************************************/
/// @public Is data weighted, or are weight factors available with data?
typedef enum {
  WEIGHTING_UNKNOWN,    ///< Not known; usually assumed that not weighted.
  WEIGHTING_OFF,        ///< Not weighted or weights not available.
  WEIGHTING_ON,         ///< Weighted or weights are available.
  // possibility to add certain weighting schemes
  WEIGHTING_LAST
} weights;
/*****************************************************************************/

/*****************************************************************************/
/// @public TAC content types
typedef enum {
  TACTYPE_UNKNOWN,      ///< Content type not known
  TACTYPE_PLASMA,       ///< Plasma TAC 
  TACTYPE_BLOOD,        ///< Blood TAC
  TACTYPE_PARENT,       ///< Parent tracer in plasma
  TACTYPE_METABOLITE,   ///< Metabolite in plasma
  TACTYPE_VOI,          ///< Image volume-of-interest TAC
  TACTYPE_REF,          ///< Reference region TAC
  TACTYPE_PIXEL,        ///< Image pixel TAC
  TACTYPE_LAST
} tactype;
/*****************************************************************************/

/*****************************************************************************/
/// @enum unit
/// @public unit code, same as unit position in the table (table can not
/// be accessed directly outside the c file).
/// @sa units.c unitName, unitIdentify, unitConversionFactor
typedef enum {
  UNIT_UNKNOWN,              ///< Unknown unit
  UNIT_UNITLESS,             ///< Unitless
  UNIT_MSEC,                 ///< milliseconds
  UNIT_SEC,                  ///< seconds
  UNIT_MIN,                  ///< minutes
  UNIT_HOUR,                 ///< hours
  UNIT_DAY,                  ///< days
  UNIT_MONTH,                ///< months
  UNIT_YEAR,                 ///< years
  UNIT_UM,                   ///< micrometer
  UNIT_MM,                   ///< millimeter
  UNIT_CM,                   ///< centimeter
  UNIT_M,                    ///< meter
  UNIT_UL,                   ///< microliter
  UNIT_ML,                   ///< milliliter
  UNIT_DL,                   ///< desiliter
  UNIT_L,                    ///< liter
  UNIT_UG,                   ///< microgram
  UNIT_MG,                   ///< milligram
  UNIT_G,                    ///< gram
  UNIT_100G,                 ///< 100 grams
  UNIT_KG,                   ///< kilogram
  UNIT_PMOL,                 ///< picomole
  UNIT_NMOL,                 ///< nanomole
  UNIT_UMOL,                 ///< micromole
  UNIT_MMOL,                 ///< millimole
  UNIT_MOL,                  ///< mole
  UNIT_COUNTS,               ///< counts
  UNIT_KCOUNTS,              ///< kilocounts
  UNIT_BQ,                   ///< Becquerel
  UNIT_KBQ,                  ///< kiloBecquerel
  UNIT_MBQ,                  ///< megaBecquerel
  UNIT_GBQ,                  ///< gigaBecquerel
  UNIT_NCI,                  ///< nanoCurie
  UNIT_UCI,                  ///< microCurie
  UNIT_MCI,                  ///< milliCurie
  UNIT_CI,                   ///< Curie
  UNIT_CPS,                  ///< counts/s
  UNIT_KCPS,                 ///< kilocounts/s
  UNIT_CPM,                  ///< counts/min
  UNIT_KCPM,                 ///< kilocounts/min
  UNIT_BQ_PER_ML,            ///< Bq/mL
  UNIT_KBQ_PER_ML,           ///< kBq/mL
  UNIT_MBQ_PER_ML,           ///< MBq/mL
  UNIT_NCI_PER_ML,           ///< nCi/mL
  UNIT_UCI_PER_ML,           ///< uCi/mL
  UNIT_BQ_PER_G,             ///< Bq/g
  UNIT_KBQ_PER_G,            ///< kBq/g
  UNIT_MBQ_PER_G,            ///< MBq/g
  UNIT_NCI_PER_G,            ///< nCi/g
  UNIT_UCI_PER_G,            ///< uCi/g
  UNIT_PMOL_PER_ML,          ///< pmol/mL
  UNIT_NMOL_PER_ML,          ///< nmol/mL
  UNIT_UMOL_PER_ML,          ///< umol/mL
  UNIT_MMOL_PER_ML,          ///< mmol/mL
  UNIT_PMOL_PER_L,           ///< pmol/L
  UNIT_NMOL_PER_L,           ///< nmol/L
  UNIT_UMOL_PER_L,           ///< umol/L
  UNIT_MMOL_PER_L,           ///< mmol/L
  UNIT_MOL_PER_L,            ///< mol/L
  UNIT_SEC_KBQ_PER_ML,       ///< s*kBq/mL
  UNIT_MIN_KBQ_PER_ML,       ///< min*kBq/mL
  UNIT_SEC_BQ_PER_ML,        ///< s*Bq/mL
  UNIT_MIN_BQ_PER_ML,        ///< min*Bq/mL
  UNIT_PERCENTAGE,           ///< Percentage (%)
  UNIT_ML_PER_ML,            ///< mL/mL
  UNIT_ML_PER_DL,            ///< mL/dL
  UNIT_G_PER_ML,             ///< g/mL
  UNIT_ML_PER_G,             ///< mL/g
  UNIT_PER_SEC,              ///< 1/s
  UNIT_PER_MIN,              ///< 1/min
  UNIT_PER_HOUR,             ///< 1/h
  UNIT_ML_PER_ML_SEC,        ///< mL/(mL*sec)
  UNIT_ML_PER_ML_MIN,        ///< mL/(mL*min)
  UNIT_ML_PER_DL_MIN,        ///< mL/(dL*min)
  UNIT_ML_PER_G_MIN,         ///< mL/(g*min)
  UNIT_ML_PER_100G_MIN,      ///< mL/(100g*min)
  UNIT_UMOL_PER_ML_MIN,      ///< umol/(mL*min)
  UNIT_UMOL_PER_DL_MIN,      ///< umol/(dL*min)
  UNIT_UMOL_PER_G_MIN,       ///< umol/(g*min)
  UNIT_UMOL_PER_100G_MIN,    ///< umol/(100g*min)
  UNIT_MMOL_PER_ML_MIN,      ///< mmol/(mL*min)
  UNIT_MMOL_PER_DL_MIN,      ///< mmol/(dL*min)
  UNIT_MMOL_PER_G_MIN,       ///< mmol/(g*min)
  UNIT_MMOL_PER_100G_MIN,    ///< mmol/(100g*min)
  UNIT_MG_PER_100G_MIN,      ///< mg/(100g*min)
  UNIT_MG_PER_DL_MIN,        ///< mg/(dL*min)
  UNIT_PID,                  ///< Percent of injected dose
  UNIT_PID_PER_G,            ///< Percent of injected dose / g
  UNIT_PID_PER_ML,           ///< Percent of injected dose / mL
  UNIT_PID_PER_KG,           ///< Percent of injected dose / kg
  UNIT_PID_PER_L,            ///< Percent of injected dose / L
  UNIT_MBQ_PER_NMOL,         ///< MBq/nmol (specific activity)
  UNIT_GBQ_PER_NMOL,         ///< GBq/nmol (specific activity)
  UNIT_LAST                  ///< end of list
} unit;
/*****************************************************************************/

/*****************************************************************************/
/// @brief Error codes.
/// @public Error code tpcerror, same as error message position in the table 
/// (table can not be accessed directly outside the c file).
/// @sa tpcerrormsg, statusmsg.c, TPCSTATUS
typedef enum {
  TPCERROR_OK,                 ///< No error
  TPCERROR_FAIL,               ///< General error
  TPCERROR_OUT_OF_MEMORY,      ///< Cannot allocate memory
  TPCERROR_NO_FILE,            ///< File does not exist
  TPCERROR_INVALID_FILENAME,   ///< Invalid file name
  TPCERROR_CANNOT_OPEN,        ///< Cannot open file
  TPCERROR_CANNOT_READ,        ///< Cannot read file
  TPCERROR_CANNOT_WRITE,       ///< Cannot write file
  TPCERROR_INVALID_FORMAT,     ///< Invalid file format
  TPCERROR_UNSUPPORTED,        ///< Unsupported file type
  TPCERROR_NO_DATA,            ///< File contains no data
  TPCERROR_DUPLICATE_DATA,     ///< File contains duplicate data
  TPCERROR_NO_WEIGHTS,         ///< File contains no weights
  TPCERROR_NO_DATETIME,        ///< File contains no date or time
  TPCERROR_MISSING_DATA,       ///< File contains missing values
  TPCERROR_TOO_FEW,            ///< File contains too few samples
  TPCERROR_TOO_BIG,            ///< File is too big
  TPCERROR_OVERLAPPING_DATA,   ///< Overlapping data
  TPCERROR_LARGE_GAP,          ///< Large gap in data
  TPCERROR_NO_X,               ///< No sample times
  TPCERROR_INVALID_X,          ///< Invalid sample time
  TPCERROR_INVALID_XRANGE,     ///< Invalid sample time range
  TPCERROR_INCOMPATIBLE_UNIT,  ///< Incompatible units
  TPCERROR_INCOMPATIBLE_DATA,  ///< Incompatible data
  TPCERROR_NO_REFERENCE,       ///< Reference not found
  TPCERROR_NO_KEY,             ///< Key not found
  TPCERROR_NO_VALUE,           ///< Value not found
  TPCERROR_INVALID_VALUE,      ///< Invalid value
  TPCERROR_INVALID_SEPARATOR,  ///< Invalid field delimiter
  TPCERROR_INVALID_PARNR,      ///< Invalid number of parameters
  TPCERROR_INVALID_HEADER,     ///< Invalid header contents
  TPCERROR_MISSING_HEADER,     ///< Missing header
  TPCERROR_UNKNOWN_ISOTOPE,    ///< Unknown isotope
  TPCERROR_UNKNOWN_UNIT,       ///< Unknown data unit
  TPCERROR_NO_SOLUTION,        ///< No solution
  TPCERROR_BAD_FIT,            ///< Fitting not successful
  TPCERROR_LAST                ///< marking the end of list
} tpcerror;
/*****************************************************************************/

/*****************************************************************************/
/** Memory holder for storing and transferring status and information
    between functions for debug and error printing purposes.

    Do not change the contents of these fields except for verbose and error;
    change other fields only using functions statusSet() and statusFree().
    @sa statusInit, statusSet, statusFree, statusPrint, tpcerror, errorMsg
 */
typedef struct TPCSTATUS {
  int verbose;         ///< Verbose level, used by statusPrint() etc.
  FILE *fp;            ///< File pointer for writing log information during development and testing.
  int forgiving;       ///< Force level, 0 for strict tests for data units etc.
  tpcerror error;      ///< Error code.
  char *current_func;  ///< Function where status was last time set.
  char *current_file;  ///< Source file name where status was last time set.
  int current_line;    ///< Source line nr where status was last time set.
  char *last_func;     ///< Function where status was previously set.
  char *last_file;     ///< Source file name where status previously set.
  int last_line;       ///< Source line nr where status was previously set.
} TPCSTATUS;
/*****************************************************************************/

/*****************************************************************************/
/* stringext */
int strTokenNr(const char *s1, const char *s2);
int strTokenNCpy(const char *s1, const char *s2, int i, char *s3, int count);
int strChrCount(const char *s1, const char *s2);
void strReplaceChar(char *s, char c1, char c2);
/////#ifndef HAVE_STRCASESTR
extern char *strcasestr(const char *haystack, const char *needle);
/////#endif
#ifndef HAVE_STRDUP
char *strdup(const char *s);
#endif
#ifndef HAVE_STRNDUP
char *strndup(const char *s, size_t n);
#endif
#ifndef HAVE_STRNLEN
size_t strnlen(const char *s, size_t n);
#else
extern size_t strnlen(const char *s, size_t n);
#endif
#ifndef HAVE_STRLCAT
size_t strlcat(char *dst, const char *src, size_t dstsize);
#else
extern size_t strlcat(char *dst, const char *src, size_t dstsize);
#endif
#ifndef HAVE_STRLCPY
size_t strlcpy(char *dst, const char *src, size_t dstsize);
#else
extern size_t strlcpy(char *dst, const char *src, size_t dstsize);
#endif
char *strstrNoQuotation(const char *haystack, const char *needle);
int strncpyCleanSpaces(char *s1, const char *s2, int maxlen);
int strCleanSpaces(char *s);
int strncpyClean(char *s1, const char *s2, int maxlen);
int strClean(char *s);
char *strTokenDup(const char *s1, const char *s2, int *next);
int strInPars(char *s);
void strCleanPars(char *s);
char *strncatInt(char *str1, const int n, size_t count);
char *strncatDouble(char *str1, const double d, size_t count);
char *strncatIntZ(char *str1, const int n, const int maxn, size_t count);
int strIsSpaceOnly(char *s);
char *strdelstr(char *s1, const char *s2);
char *strTrimLeft(char *s, const size_t t);
/*****************************************************************************/

/*****************************************************************************/
/* decpoint */
int strIsValidNumber(const char *s);
double atofVerified(const char *s);
int atofCheck(const char *s, double *v);
int strHaveDecimalComma(const char *s);
int strHaveDecimalSeparator(const char *s);
int atofList(const char *s1, const char *s2, double *x, int maxn);
/*****************************************************************************/

/*****************************************************************************/
/* filename */
void filenameRmPath(char *s);
void filenameRmFile(char *s);
int filenameRmExtension(char *s);
void filenameRmExtensions(char *s);
int fnmatch(const char *fname, const char *key);
int fncasematch(const char *fname, const char *key);
char *filenameGetExtension(const char *s);
char *filenameGetExtensions(const char *s);
/*****************************************************************************/

/*****************************************************************************/
/* proginfo */
int tpcProcessStdOptions(const char *s, int *print_usage, int *print_version, int *verbose_level);
void tpcPrintUsage(const char *program, char *text[], FILE *fp);
int tpcHtmlUsage(const char *program, char *text[], const char *path);
void tpcPrintBuild(const char *program, FILE *fp);
void tpcProgramName(const char *program, int version, int copyright, char *prname, int n);
int tpcYesNo(const char *s);
/*****************************************************************************/

/*****************************************************************************/
/* units */
char *unitName(int unit_code);
int unitIdentify(const char *s);
int unitIdentifyFilename(const char *s);
int unitIsDistance(int u);
int unitIsTime(int u);
int unitIsVolume(int u);
int unitIsMass(int u);
int unitIsMole(int u);
int unitIsRadioactivity(int u);
int unitIsRAConc(int u);
int unitIsCombinatorial(int u);
double unitConversionFactor(const int u1, const int u2);
int unitInverse(int u);
int unitDividerHasVolume(int u);
int unitDividerHasMass(int u);
int unitDividendHasRadioactivity(int u);
int unitCombination(const int u1, const int u2, const int v1, const int v2);
int unitDividerMassVolumeConversion(int u);
/*****************************************************************************/

/*****************************************************************************/
/* statusmsg */
void statusPrint(FILE *fp, TPCSTATUS *s);
void statusInit(TPCSTATUS *s);
void statusFree(TPCSTATUS *s);
void statusSet(TPCSTATUS *s, const char *func, const char *srcfile, int srcline, tpcerror error);
char *errorMsg(tpcerror e);
/*****************************************************************************/

/*****************************************************************************/
/* readasciifile */
size_t asciiFileSize(FILE *fp, int *nonprintable);
char *asciiFileRead(FILE *fp, char *data, size_t maxlen);
int asciiCommentLine(const char *line, int *cont);
/*****************************************************************************/

/*****************************************************************************/
/* datetime */
#ifndef HAVE_GMTIME_R
struct tm* gmtime_r(const time_t* t, struct tm* tm);
#endif
#ifndef HAVE_LOCALTIME_R
struct tm* localtime_r(const time_t* t, struct tm* tm);
#endif
#ifndef HAVE_TIMEGM
time_t timegm(struct tm *tm);
#else
extern time_t timegm(struct tm *tm); // needed at least in OSX
#endif
char* ctime_r_int(const time_t *t, char *buf);
int strDateValid(const char *str);
int strDateValid2(const char *str, char *intdate);
int strDateValid3(const char *str, char *intdate);
int strDateValid4(int dateint, char *intdate, int *year, int *month, int *day);
int strTimeValid(const char *str);
int strDateTimeValid(const char *str, char *intdate);
int strDateTimeRead(const char *str, struct tm *date);
int strDateRead(const char *str, struct tm *date);
void time_to_tm(time_t totalsecs, int offset, struct tm *result);
double tmDifference(struct tm *tm1, struct tm *tm0);
void tmAdd(int s, struct tm *d);
double strDateTimeDifference(const char *dt1, const char *dt0);
int strDateTimeAdd(int s, char *dt);
/*****************************************************************************/

/*****************************************************************************/
/* studynr */
int studynrVerify(const char *s, int zero_ok);
int studynrStandardize(char *s);
int studynrFromFilename(const char *fname, char *studynr, int force);
/*****************************************************************************/

/*****************************************************************************/
/* doubleutil */
int doubleMatch(const double v1, const double v2, const double lim);
int doubleArrayMatch(const double *a1, const double *a2, const unsigned int n, const double lim);
int doubleMatchRel(const double v1, const double v2, const double lim);
double doubleMachEps();
void doubleCopy(double *t, double *s, const unsigned int n);
unsigned int doubleCopyFinite(double *t, double *s, const unsigned int n);
unsigned int doubleNaNs(double *a, const unsigned int n);
unsigned int doubleRange(double *a, const unsigned int n, double *amin, double *amax);
double doubleSum(double *a, const unsigned int n);
double doubleMean(double *a, const unsigned int n);
double doubleWMean(double *a, double *w, const unsigned int n);
int doubleGetWithUnit(const char *s, double *v, int *u);
int doubleSpanPositives(double *a, const int n);
int doubleCSpanPositives(double *a, const int n);
unsigned int doubleNonzeroes(double *a, const unsigned int n);
unsigned int doubleMaxIndex(double *a, const unsigned int n);
unsigned int doubleAbsMaxIndex(double *a, const unsigned int n);
unsigned int doubleMinIndex(double *a, const unsigned int n);
unsigned int doubleAbsMinIndex(double *a, const unsigned int n);
unsigned int doubleGEIndex(double *a, const unsigned int n, double lim);
unsigned int doubleGTIndex(double *a, const unsigned int n, double lim);
double inverfc(double x);
/*****************************************************************************/

/*****************************************************************************/
/* floatutil */
int floatMatch(const float v1, const float v2, const float lim);
int floatMatchRel(const float v1, const float v2, const float lim);
float floatMachEps();
void floatCopy(float *t, float *s, const unsigned int n);
unsigned int floatMaxIndex(float *a, const unsigned int n);
float floatSum(float *a, const unsigned int n);
float floatMean(float *a, const unsigned int n);
int floatGetWithUnit(const char *s, float *v, int *u);
int floatSpanPositives(float *a, const int n);
int floatCSpanPositives(float *a, const int n);
unsigned int floatNonzeroes(float *a, const unsigned int n);
/*****************************************************************************/

/*****************************************************************************/
/* intutil */
int atoiCheck(const char *s, int *v);
unsigned int factorial(unsigned int n);

/** List of integer values.
    @sa integerListInit, integerListEmpty, integerListAdd, integerListSort. */
typedef struct INTLIST {
  /** Nr of stored integers. */
  int nr;
  /** Allocated list size, >= nr. */
  int _nr;
  /** Array of integers. */
  int *i;
} INTLIST;
void intlistInit(INTLIST *l);
void intlistFree(INTLIST *l);
int intlistAdd(INTLIST *l, const int v, const int ifnew);
void intlistSort(INTLIST *l);
int intlistAddFromString(const char *s1, const char *s2, INTLIST *l, const int ifnew);
int intlistExpandFromString(const char *s1, const char *s2, INTLIST *l, const int ifnew);
/*****************************************************************************/

/*****************************************************************************/
/* roiname */
char *roinameSubpart(
  const char *roiname, const char *dlm, const unsigned int si, 
  char *subpart, const unsigned int slen
);
char *roinameEditByTemplate(
  const char *template, const char *currname,
  char *newname, const unsigned int count 
);
char *roinameAddField(
  char *roiname, const char *field, 
  const unsigned int in, const unsigned int count
);
int roinameExists(char *roiname);
int roinameMatch(const char *roiname, const char *test_str, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* endian */
int endianLittle();
void swap(void *from, void *to, int size);
void swabip(void *buf, int size);
void swawbip(void *buf, int size);
void swawip(void *buf, int size);
/*****************************************************************************/

/*****************************************************************************/
#endif /* TPCEXTENSIONS */
