/** @file tpcnlopt.h
 *  @brief Header file for library libtpcnlopt.
 *  @details Non-linear optimization.
 *  @author Vesa Oikonen
 */
#ifndef _TPCNLOPT_H_
#define _TPCNLOPT_H_
/*****************************************************************************/

/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
/*****************************************************************************/
#include "tpcextensions.h"
#include "tpcrand.h"
/*****************************************************************************/

/*****************************************************************************/
/** General settings for NLOPT routines; individual routines may not use all of these,
    and can require additional parameters. */
typedef struct NLOPT {
  /** Total nr of parameters (dimensions). */
  unsigned int totalNr;
  /** Array of all parameter values, including fixed ones. */
  double       *xfull;
  /** Array of lower limits for parameters, including fixed ones. */
  double       *xlower;
  /** Array of upper limits for parameters, including fixed ones. */
  double       *xupper;
  /** Array of (initial) steps for parameters, including fixed ones;
      set it to the expected distance from the solution. */
  double       *xdelta;
  /** Array of tolerances for parameters, including fixed ones; may be used
      as a stopping rule and for scaling. */
  double       *xtol;
  /** Pointer to objective function; parameters are optimized to minimize
      the return value of this function. */
  double (*_fun)(int, double*, void*);
  /** Pointer to data which is passed on to the function; NULL if not needed. */
  void *fundata;
  /** Maximum allowed number of function calls; set to zero for no limit. */
  unsigned int maxFunCalls;
  /** Number of function calls so far. */
  unsigned int funCalls;
  /** The return value of objective function with xfull[] and fundata. */
  double funval;
  /** Keep (1) or do not keep (0) plist (below). */
  unsigned int usePList;
  /** List of all visited parameters and objective function values;
      stored in order x[], funval, x[], funval, ... with length funCalls*(totalNr+1);
      updated only if usePList!=0 (above). */
  double       *plist;
} NLOPT;
/*****************************************************************************/

/*****************************************************************************/
/** Basic structure for data passed on to objective function via NLOPT structure.
    Actual implementations usually require more data, and instead of this structure,
    a dedicated local data structure should be created and used.
 */
typedef struct NLOPT_DATA {
  /** Sample number. */
  unsigned int  n;
  /** Pointer to an array of sample times. */
  double       *x;
  /** Pointer to an array of sample values. */
  double       *y;
  /** Pointer to an array of weights. */
  double       *w;
  /** Pointer to fitted/simulated sample values. */
  double       *sy;
  /** Verbose level. */
  int           verbose;
  /** Pointer to verbose output; usually stdout. Must be set if verbose>0. */
  FILE         *fp;
} NLOPT_DATA;
/*****************************************************************************/

/*****************************************************************************/
/* nlopt */
void nloptInit(NLOPT *nlo);
void nloptFree(NLOPT *nlo);
int nloptAllocate(NLOPT *nlo, unsigned int parNr);
int nloptDuplicate(NLOPT *nlo1, NLOPT *nlo2);
int nloptAddP(NLOPT *nlo, double *p, double funval);
int nloptSortP(NLOPT *nlo);
int nloptMeanP(NLOPT *nlo, unsigned int nr, double *meanp, double *sdp);
void nloptPrintP(NLOPT *nlo, unsigned int nr, FILE *fp);
void nloptWrite(NLOPT *d, FILE *fp);
unsigned int nloptLimitFixedNr(NLOPT *d);
unsigned int nloptFixedNr(NLOPT *d);
void nloptRemoveEmpties(NLOPT *d);
void nloptdataInit(NLOPT_DATA *d);
/*****************************************************************************/

/*****************************************************************************/
/* rndpoint */
int nloptRandomPoint(double *p, double *low, double *up, unsigned int n, MERTWI *mt);
int nloptGaussianPoint(
  double *p, double *mean, double *sd, double *low, double *up, unsigned int n, MERTWI *mt
);
/*****************************************************************************/

/*****************************************************************************/
/* constraints */
unsigned int nloptCheckParameters(
  unsigned int n, double *pLower, double *pUpper, double *p,
  double *pAccept, double *penalty
);
unsigned int nloptForceLimits(unsigned int n, double *pLower, double *pUpper, double *p);
/*****************************************************************************/

/*****************************************************************************/
/* nlopt1d */
int nlopt1D(NLOPT *nlo, unsigned int maxIter, TPCSTATUS *);
/*****************************************************************************/

/*****************************************************************************/
/* simplex */
int nloptSimplex(NLOPT *nlo, unsigned int maxIter, TPCSTATUS *);
int nloptSimplexARRS(NLOPT *nlo, unsigned int maxIter, TPCSTATUS *status);
/*****************************************************************************/

/*****************************************************************************/
/* praxis */
int nloptPowellBrent(NLOPT *nlo, unsigned int ktm, TPCSTATUS *);
/*****************************************************************************/

/*****************************************************************************/
/* mpso */
int nloptMPSO(
  NLOPT *nlo, unsigned int maxIter, unsigned int nSwarms, unsigned int nParticles, double wInertia,
  double wParticle, double wSwarm, double wGlobal, double pDeath, double pImmigration,
  const int doLocal, TPCSTATUS *
);
/*****************************************************************************/

/*****************************************************************************/
/* tgo */
int nloptITGO1(
  NLOPT *nlo, const int doLocal, unsigned int tgoNr, unsigned int sampleNr,
  unsigned int neighNr, TPCSTATUS *status
);
int nloptITGO2(
  NLOPT *nlo, const int doLocal, unsigned int tgoNr, unsigned int sampleNr,
  unsigned int neighNr, TPCSTATUS *status
);
int nloptIATGO(NLOPT *nlo, const int doLocal, unsigned int maxIterNr,
  double neighFract, TPCSTATUS *status
);
/*****************************************************************************/

/*****************************************************************************/
#endif /* TPCNLOPT */
