/** @file tacw.c
 *  @brief Working with weights in TAC structure.
 */
/*****************************************************************************/
#include "tpcclibConfig.h"
/*****************************************************************************/
#include "tpcift.h"
/*****************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*****************************************************************************/
#include "tpctac.h"
/*****************************************************************************/

/*****************************************************************************/
/** Copy weights from one TAC structure to another.
    @sa tacXCopy, tacWMove, sifWeight, tacWByFreq
    @author Vesa Oikonen
    @return Returns TPCERROR status.
 */
int tacWCopy(
  /** Pointer to source TAC structure. */
  TAC *tac1,
  /** Pointer to target TAC structure. */
  TAC *tac2,
  /** Index of the first sample to copy. */
  int i1,
  /** Index of the last sample to copy. */
  int i2
) {
  if(tac1==NULL || tac2==NULL) return(TPCERROR_FAIL);
  if(i1<0 || i1>i2) return(TPCERROR_FAIL);
  if(i2>=tac1->_sampleNr || i2>=tac2->_sampleNr) return(TPCERROR_FAIL);
  
  for(int i=i1; i<=i2; i++) {
    tac2->w[i]=tac1->w[i];
  }
  tac2->weighting=tac1->weighting;
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Identify column containing weights in TAC structure, and move weights to the correct place. 

    TAC name starting with string 'weight' is identified as weight column.

    Sets tac->weighting to WEIGHTING_ON or WEIGHTING_OFF, depending on whether
    weight column was found or not.
  
    @sa tacWCopy
    @author Vesa Oikonen
    @return enum tpcerror (TPCERROR_OK when successful).
 */
int tacWMove(
  /** Pointer to TAC structure. */
  TAC *tac,
  /** Overwrite (1) or do not overwrite (0) existing weights. */
  int ow,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  int ret;

  if(tac->weighting==WEIGHTING_UNKNOWN) tac->weighting=WEIGHTING_OFF;

  /* Search for the weight column */
  int wc=-1;
  for(int i=0; i<tac->tacNr; i++) {
    if(strncasecmp(tac->c[i].name, "WEIGHT", 6)!=0) continue;
    if(wc<0) wc=i;
    else if(ow==0) {
      /* Error if found more than one */
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_DUPLICATE_DATA);
      return TPCERROR_DUPLICATE_DATA;
    }
  }
  /* Error if not found */
  if(wc<0) {
    if(ow) tac->weighting=WEIGHTING_OFF;
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_WEIGHTS);
    return TPCERROR_NO_WEIGHTS;
  }
  /* Error if overwriting not allowed */
  if(ow==0 && tac->weighting==WEIGHTING_ON) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_DUPLICATE_DATA);
    return TPCERROR_DUPLICATE_DATA;
  }
  /* Copy weights to their right place */
  for(int i=0; i<tac->sampleNr; i++) tac->w[i]=tac->c[wc].y[i];
  tac->weighting=WEIGHTING_ON;
  /* Then delete the original column */
  ret=tacDeleteTACC(tac, wc);
  statusSet(status, __func__, __FILE__, __LINE__, ret);
  return(ret);
}
/*****************************************************************************/

/*****************************************************************************/
/** Add weight to TAC data based on sample frequency or time frame length.

    Weights are scaled so that weight sum = sample number.
    Any existing weighting is overwritten.

    @sa sifWeight,tacWSampleNr, tacWCopy, tacWMove, tacWeightModerate
    @return Return enum tpcerror (TPCERROR_OK when successful).
 */
int tacWByFreq(
  /** Pointer to TAC struct where weights will be added.
      @note Samples/frames must be sorted by sample time, but duplicate samples are allowed. */
  TAC *tac,
  /** Isotope code, in case the weights are calculated for decay corrected data.
      Enter ISOTOPE_UNKNOWN if calculating weights for non-corrected data, or when
      decay correction is ignored. */
  isotope isot,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s(%s)\n", __func__, isotopeName(isot));
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  
  tac->weighting=WEIGHTING_ON;
  if(tac->sampleNr==1) {tac->w[0]=1.0; return(TPCERROR_OK);}

  if(tac->isframe!=0) { /* weights based on frame lengths */

    for(int i=0; i<tac->sampleNr; i++) {
      tac->w[i]=tac->x2[i]-tac->x1[i]; 
    }

  } else { /* weights based on sample distance */

    int i, i1, i2;
    double t, t1, t2, f;
    for(i=0; i<tac->sampleNr; i++) {
      t=t1=t2=tac->x[i];
      /* Find the closest sample time before this one */
      for(i1=i; i1>=0; i1--) {t1=tac->x[i1]; if(t1<t) break;} 
      /* Find the closest sample time after this one */ 
      for(i2=i; i2<tac->sampleNr; i2++) {t2=tac->x[i2]; if(t2>t) break;} 
      /* Mean sample distance */
      f=0.0;
      if(t1<t) f+=t-t1; else f+=t2-t;
      if(t2>t) f+=t2-t; else f+=t-t1;
      f*=0.5; if(f<=0.0) f=1.0;
      tac->w[i]=f; 
    }

  }

  /* Account for decay correction, if requested */
  if(isot!=ISOTOPE_UNKNOWN) {
    double lambda=lambdaFromIsotope(isot);
    if(tac->tunit==UNIT_SEC) lambda/=60.; else if(tac->tunit==UNIT_MSEC) lambda/=60000.;
    if(tac->isframe!=0) {
      for(int i=0; i<tac->sampleNr; i++)
        tac->w[i]*=decayCorrectionFactorFromLambda(-lambda, tac->x1[i], (tac->x2[i]-tac->x1[i])); 
    } else {
      for(int i=0; i<tac->sampleNr; i++)
        tac->w[i]*=decayCorrectionFactorFromLambda(-lambda, tac->x[i], 0.0); 
    }
  }

  /* Scale weights so that sum of weights equals sample number */
  double sumw=0.0;
  for(int i=0; i<tac->sampleNr; i++) if(isfinite(tac->w[i])) sumw+=tac->w[i];
  if(!(sumw>0.0)) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_VALUE);
    return TPCERROR_INVALID_VALUE;
  }
  sumw/=(double)tac->sampleNr;
  for(int i=0; i<tac->sampleNr; i++) tac->w[i]/=sumw;

  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
/** Get the number of samples in TAC that have weight > 0.
    Missing (NaN) sample values are included as long as weight is not missing.
    If weights are not set, then nr of all samples is returned.
   @return The number of samples with weight above zero.
   @sa tacWByFreq, aicSS, parFreeNr, tacWeightModerate
*/
unsigned int tacWSampleNr(
  /** Pointer to the TAC struct. */
  TAC *tac
) {
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) return((unsigned int)0);
  if(tac->weighting==WEIGHTING_OFF) return((unsigned int)tac->sampleNr);
  unsigned int n=0;
  for(int i=0; i<tac->sampleNr; i++) if(tac->w[i]>0.0) n++;
  if(n==0 && tac->weighting==WEIGHTING_UNKNOWN) n=(unsigned int)tac->sampleNr;
  return(n);
}
/*****************************************************************************/

/*****************************************************************************/
/** Scale weights so that average weight is 1.0, and the sum of weights equals the nr of frames.
   @sa tacWeightModerate,sifWeight, tacWCopy, tacWByFreq, tacWSampleNr, tacReadSIF
   @return enum tpcerror (TPCERROR_OK when successful).
   @todo Add tests
*/
int tacWeightNorm(
  /** Pointer to TAC structure (which can contain SIF or TAC data). */
  TAC *tac,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s()\n", __func__);
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  if(tac->weighting!=WEIGHTING_ON) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_WEIGHTS);
    return TPCERROR_NO_WEIGHTS;
  }

  double wSum=0.0; int wNr=0;
  for(int i=0; i<tac->sampleNr; i++)
    if(isfinite(tac->w[i]) && tac->w[i]>0.0) {wSum+=tac->w[i]; wNr++;}
  if(wNr==0 || !(wSum>0.0)) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_VALUE);
    return TPCERROR_INVALID_VALUE;
  }
  double wMean=wSum/(double)wNr;
  for(int i=0; i<tac->sampleNr; i++) 
    if(isfinite(tac->w[i]) && tac->w[i]>0.0) tac->w[i]/=wMean;

  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return TPCERROR_OK;
}
/*****************************************************************************/

/*****************************************************************************/
/** Moderate weights so that the min weight is not less than given proportion of the maximum weight.

    After moderation, all weights are normalized to have and average of one.
    @sa sifWeight, tacWeighNorm, tacWCopy, tacWByFreq, tacWSampleNr, tacReadSIF
    @return enum tpcerror (TPCERROR_OK when successful).
    @todo Add tests
*/
int tacWeightModerate(
  /** Pointer to TAC structure (which can contain SIF or TAC data). */
  TAC *tac,
  /** The minimum proportion of frame weight to maximum frame weight. Must be less than one.
      If proportion is lower than this limit, then weight is set to proportion times maximum. 
      Setting minimum proportion to zero or less does not change any weights, but weights are
      still normalized. By default, zero weights or NaNs are not changed. */
  const double minprop,
  /** Set zero weights, too, to the minimum proportion of maximum frame weight. */
  const int doZeroes,
  /** Set NaN weights, too, to the minimum proportion of maximum frame weight. */
  const int doNaNs,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s(%g, %d, %d)\n", __func__, minprop, doZeroes, doNaNs);
  if(tac==NULL || tac->tacNr<1 || tac->sampleNr<1) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  if(!isfinite(minprop) || minprop>=1.0) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_VALUE);
    return TPCERROR_INVALID_VALUE;
  }
  if(tac->weighting!=WEIGHTING_ON) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_WEIGHTS);
    return TPCERROR_NO_WEIGHTS;
  }

  double wMax=nan("");
  for(int i=0; i<tac->sampleNr; i++) if(isfinite(tac->w[i])) {
    if(!isfinite(wMax)) {wMax=tac->w[i]; continue;}
    if(tac->w[i]>wMax) wMax=tac->w[i];
  }
  if(verbose>2) {printf("weight_maximum := %g\n", wMax); fflush(stdout);}
  if(!isfinite(wMax) || wMax<=0.0) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_WEIGHTS);
    return TPCERROR_NO_WEIGHTS;
  }

  double wLimit=minprop*wMax;
  if(verbose>2) {printf("weight_limit := %g\n", wLimit); fflush(stdout);}

  int n=0;
  if(wLimit>0.0) {
    for(int i=0; i<tac->sampleNr; i++) {
      if(tac->w[i]==0.0) {
        if(doZeroes) {tac->w[i]=wLimit; n++;}
        continue;
      }
      if(isnan(tac->w[i])) {
        if(doNaNs) {tac->w[i]=wLimit; n++;}
        continue;
      }
      if(tac->w[i]<wLimit) {tac->w[i]=wLimit; n++;}
    }
  }
  if(verbose>1 && minprop>0.0) {printf("  %d weight(s) set to %g.\n", n, wLimit); fflush(stdout);}

  return(tacWeightNorm(tac, status));
}
/*****************************************************************************/

/*****************************************************************************/
/** Calculate weights for frames in SIF data based on true counts. 

    Weights are calculated from formula weight=(frame duration)^2 / (trues in a frame)
    which is applicable to non-decay corrected data.
    If weights are to be applied to decay-corrected data, provide function with isotope to
    use formula weight=(frame duration)^2 / (decay corrected trues in a frame) instead.

    Reference: Mazoyer BM, Huesman RH, Budinger TF, Knittel BL. 
    Dynamic PET data analysis. J Comput Assist Tomogr 1986; 10:645-653.

    Formula is not applicable to initial frames where trues can be zero or even less,
    but weights for frames with no trues should still be high for fitting purposes;
    therefore, in those cases, weight is set to the maximum weight that was calculated
    from the data.

    Weights are normalized to have an average of 1.0 for frames that have weight above zero.

    @sa tacWCopy, tacWByFreq, tacWSampleNr, tacReadSIF, tacWeightModerate, tacWeightNorm
    @return enum tpcerror (TPCERROR_OK when successful).
    @todo Add tests

 */
int sifWeight(
  /** Pointer to SIF data, stored in TAC structure. 
      Weights are added to the structure.
      Data must have at least two columns (TACs), which are assumed to represent
      prompts and randoms, to calculate trues as prompts - randoms.
      Counts in SIF are assumed to be not corrected for physical decay. 
      Frame times are assumed to be in seconds. */
  TAC *sif,
  /** Isotope code, in case the weights are calculated for decay corrected data.
      Enter ISOTOPE_UNKNOWN if calculating weights for non-corrected data. 
      SIF should contain isotope, too, but that is not used in this function. */
  isotope isot,
  /** Pointer to status data; enter NULL if not needed */
  TPCSTATUS *status
) {
  int verbose=0; if(status!=NULL) verbose=status->verbose;
  if(verbose>0) printf("%s(%s)\n", __func__, isotopeName(isot));
  if(sif==NULL || sif->tacNr<1 || sif->sampleNr<1) {
    if(sif!=NULL && verbose>2) {
      printf("  sif.tacNr := %d\n  sif.sampleNr := %d\n", sif->tacNr, sif->sampleNr);
    }
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_NO_DATA);
    return TPCERROR_NO_DATA;
  }
  sif->weighting=WEIGHTING_OFF;
  if(sif->tacNr<2 || !sif->isframe) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_UNSUPPORTED);
    return TPCERROR_UNSUPPORTED;
  }
  /* Check times */
  double xmin, xmax;
  if(tacXRange(sif, &xmin, &xmax) || !(xmax>xmin)) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_XRANGE);
    return TPCERROR_INVALID_XRANGE;
  }

  /* Calculate trues */
  double trues[sif->sampleNr];
  for(int i=0; i<sif->sampleNr; i++) trues[i]=sif->c[0].y[i]-sif->c[1].y[i];
  if(verbose>2) {
    printf("\nTrues:\n");
    for(int i=0; i<sif->sampleNr; i++) printf("\t%g\n", trues[i]);
    fflush(stdout);
  }
  /* Correct for decay, if requested */
  if(isot!=ISOTOPE_UNKNOWN) {
    for(int i=0; i<sif->sampleNr; i++)
      trues[i]*=decayCorrectionFactorFromIsotope(isot, sif->x1[i]/60., (sif->x2[i]-sif->x1[i])/60.);
    if(verbose>2) {
      printf("\nTrues after decay correction:\n");
      for(int i=0; i<sif->sampleNr; i++) printf("\t%g\n", trues[i]);
      fflush(stdout);
    }
  }
  /* Calculate weights */
  double wMax=0.0;
  for(int i=0; i<sif->sampleNr; i++) {
    double fdur=sif->x2[i]-sif->x1[i]; if(fdur<0.0) {
      statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_X);
      return TPCERROR_INVALID_X;
    }
    if(trues[i]>0.0) {
      sif->w[i]=fdur*fdur/trues[i];
      if(sif->w[i]>wMax) wMax=sif->w[i];
    } else {
      sif->w[i]=nan("");
    }
  }
  if(!(wMax>0.0)) {
    statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_INVALID_VALUE);
    return TPCERROR_INVALID_VALUE;
  }
  if(verbose>1) {printf("\twMax := %g\n", wMax); fflush(stdout);}
  /* Set maximum weight to frames which had no true counts */
  for(int i=0; i<sif->sampleNr; i++) if(!(trues[i]>0.0) && isnan(sif->w[i])) sif->w[i]=wMax;
  if(verbose>2) {
    printf("\nWeights before normalization:\n");
    for(int i=0; i<sif->sampleNr; i++) printf("\t%g\n", sif->w[i]);
    fflush(stdout);
  }

  /* Normalize weights */
  sif->weighting=WEIGHTING_ON;
  if(tacWeightNorm(sif, status)!=TPCERROR_OK) return(status->error);

  statusSet(status, __func__, __FILE__, __LINE__, TPCERROR_OK);
  return(TPCERROR_OK);
}
/*****************************************************************************/

/*****************************************************************************/
